// Copyright (C) 2002 Neil Stevens <neil@hakubi.us>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
// AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name(s) of the author(s) shall not be
// used in advertising or otherwise to promote the sale, use or other dealings
// in this Software without prior written authorization from the author(s).

#include <cassert>
#include <kapplication.h>
#include <kdebug.h>
#include <kfilemetainfo.h>
#include <klocale.h>
#include <noatun/app.h>
#include <noatun/player.h>
#include <qheader.h>
#include <qvaluestack.h>
#include <qtimer.h>

#include "playlist.h"


namespace
{
QString relativeString(const KURL& parent, const KURL& child)
{
	QString childString = child.url();
	QString parentString = parent.url(1);
	childString.remove(0, parentString.length());
	return KURL::decode_string(childString);
}
}

//////////////////////
// PlaylistItemData //
//////////////////////

Hayes::PlaylistItemData::PlaylistItemData(const KFileItem &i)
	: url(i.url())
	, item(new KFileItem(i))
	, len(-1)
	, dict(0)
{
}

Hayes::PlaylistItemData::~PlaylistItemData(void)
{
	delete item;
	item = 0;
	delete dict;
	dict = 0;
}

QString Hayes::PlaylistItemData::property(const QString &key, const QString &) const
{
	QString capKey = key;
	capKey.replace(0, 1, capKey[0].upper());
	QString lowKey = key;
	lowKey.replace(0, 1, capKey[0].lower());

	// Ideally url would be the only property not handled by KFileMetaInfo.
	// index is there for one reason alone: the plugin called Flood.
	if(key == "url" || key == "index")
	{
		return url.prettyURL();
	}
	// Length is treated specially becuase the KFileMetaInfo is often
	// inaccurate, at least in my experience.  And noatun's value is accurate
	// by definition.
	else if(key == "length")
	{
		if(len != -1 || !item || !item->metaInfo().isValid())
			return QString::number(len);

		int length = item->metaInfo().value(key).toInt();
		return QString::number(length * 1000);
	}
	else if(item && item->metaInfo().contains(key))
	{
		return item->metaInfo().value(key).toString();
	}
	else if(item && item->metaInfo().contains(capKey))
	{
		return item->metaInfo().value(capKey).toString();
	}
	else if(item && item->metaInfo().contains(lowKey))
	{
		return item->metaInfo().value(lowKey).toString();
	}
	else
	{
		// fallback on NFMI
		if(dict)
		{
			QString *retval = dict->find(key);
			if(retval) return *retval;
		}

		// Not found
		return QString::null;
	}
}

void Hayes::PlaylistItemData::setProperty(const QString &key, const QString &value)
{
	if(key == "length")
	{
		len = value.toInt();
	}
	else
	{
		if(!dict)
		{
			dict = new QDict<QString>;
			dict->setAutoDelete(true);
		}
		dict->replace(key, new QString(value));
	}
}

void Hayes::PlaylistItemData::clearProperty(const QString &key)
{
	if(dict)
	{
		dict->remove(key);
	}
}

QStringList Hayes::PlaylistItemData::properties(void) const
{
	QStringList list;
	list.append("url");
	list.append("index");
	if(item && item->metaInfo().isValid())
		list += item->metaInfo().supportedKeys();
	if(!list.contains("length")) list.append("length");
	return list;
}

bool Hayes::PlaylistItemData::isProperty(const QString &key) const
{
	QString capKey = key;
	capKey.replace(0, 1, capKey[0].upper());
	QString lowKey = key;
	lowKey.replace(0, 1, lowKey[0].upper());
	return key == "url" ||
	       key == "index" ||
	       key == "length" ||
	       (item && item->metaInfo().isValid() && item->metaInfo().contains(key)) ||
	       (item && item->metaInfo().isValid() && item->metaInfo().contains(capKey)) ||
	       (item && item->metaInfo().isValid() && item->metaInfo().contains(lowKey)) ||
	       (dict && dict->find(key))
	       ;
}

bool Hayes::PlaylistItemData::operator==(const PlaylistItemData &b) const
{
	return url == b.url;
}

const KFileItem &Hayes::PlaylistItemData::fileItem(void) const
{
	return *item;
}

//////////////
// Playlist //
//////////////

Hayes::Playlist::Playlist(QWidget *viewParent, QWidget *parent, const char *viewName, const char *name)
	: ::Playlist(parent, name)
	, DCOPObject("Hayes")
	, treeView(new FileTreeView(viewParent, viewName))
	, myBranch(0)
	, currentItem(0)
	, itemToOpen(0)
	, saveVolume(false)
	, shuffle(false)
	, historyPosition(history.end())
{
	napp->setAutoPlay(false);
	connect(treeView, SIGNAL(executed(QListViewItem *)), this, SLOT(executed(QListViewItem *)));
	connect(treeView, SIGNAL(itemTaken(FileTreeViewItem *)), this, SLOT(itemDeleted(FileTreeViewItem *)));
	connect(treeView, SIGNAL(clearAndReopen(void)), this, SLOT(clearAndReopen(void)));
}

void Hayes::Playlist::open(const KURL &newRoot)
{
	if(newRoot == root) return;
	clear();
	myBranch = new Branch(treeView, newRoot, newRoot.prettyURL());
	connect(myBranch, SIGNAL(clear(void)), this, SLOT(cleared(void)));
	treeView->addBranch(myBranch);
	root = newRoot;
	myBranch->root()->setOpen(true);
}

KURL Hayes::Playlist::playlistDirectory(void) const
{
	return root;
}

void Hayes::Playlist::clearAndReopen(void)
{
	KURL savedItem;
	if(currentItem)
		savedItem = currentItem->url();

	KURL savedRoot = root;
	clear();
	open(savedRoot);

	if(savedItem != KURL())
		setCurrent(savedItem);
}

void Hayes::Playlist::clear(void)
{
	//kdDebug(66666) << "clear" << endl;
	setCurrentItem(0);
	if(myBranch) treeView->removeBranch(myBranch);
	history.clear();
	historyPosition = history.end();
	root = KURL();
}

void Hayes::Playlist::setShuffle(bool b)
{
	if(shuffle != b)
	{
		shuffle = b;
		emit shuffleChanged(b);
	}
}

void Hayes::Playlist::toggleShuffle(void)
{
	setShuffle(!shuffle);
}

bool Hayes::Playlist::isShuffling(void) const
{
	return shuffle;
}

void Hayes::Playlist::setSaveVolume(bool b)
{
	if(saveVolume != b)
	{
		saveVolume = b;
		emit saveVolumeChanged(b);
	}
}

void Hayes::Playlist::toggleSaveVolume(void)
{
	setSaveVolume(!saveVolume);
}

bool Hayes::Playlist::isSavingVolume(void) const
{
	return saveVolume;
}

void Hayes::Playlist::reset(void)
{
	if(!myBranch) return;
	setCurrentItem(getFirstItem(true, false));
}

void Hayes::Playlist::setCurrent(const PlaylistItem &givenItem)
{
	if(!myBranch) return;
	FileTreeViewItem *item = findItem(givenItem);
	if(!item) return;

	setCurrentItem(item);
	playCurrent();
}

void Hayes::Playlist::setCurrent(const KURL &url)
{
	setCurrent(makePlaylistItem(findItem(url)));
}

void Hayes::Playlist::setCurrent(const QString &url)
{
	setCurrent(KURL(url));
}

PlaylistItem Hayes::Playlist::next(void)
{
	//kdDebug(66666) << "next" << endl;
	if(!myBranch) return 0;
	setCurrentItem(getNextItem(currentItem, true, true));
	playCurrent();
	return current();
}

PlaylistItem Hayes::Playlist::nextSection(void)
{
	if(!myBranch) return 0;
	if(!currentItem) return next();
	QListViewItem *currentSection = currentItem->parent();
	FileTreeViewItem *item = currentItem;
	while(item && item->parent() == currentSection)
		item = getNextItem(item, true, true);
	setCurrentItem(item);
	playCurrent();
	return current();
}

PlaylistItem Hayes::Playlist::current(void)
{
	if(!myBranch) return 0;
	if(!currentItem)
	{
		FileTreeViewItem *item = getFirstItem(true, true);
		if(item) setCurrentItem(item);
	}
	return makePlaylistItem(currentItem);
}

PlaylistItem Hayes::Playlist::previous(void)
{
	if(!myBranch) return 0;
	setCurrentItem(getPreviousItem(currentItem, true, true));
	if(!currentItem) reset();
	playCurrent();
	return current();
}

PlaylistItem Hayes::Playlist::previousSection(void)
{
	if(!myBranch) return 0;
	if(!currentItem) return previous();

	// Find last playable item in the previous section
	QListViewItem *currentSection = currentItem->parent();
	FileTreeViewItem *item = currentItem;
	while(item && item->parent() == currentSection)
		item = getPreviousItem(item, true, true);

	// Find first playable item in the previous section
	FileTreeViewItem *lastItem = item;
	if(item)
	{
		currentSection = item->parent();
		do
		{
			if(item->isOn())
				lastItem = item;
			item = static_cast<FileTreeViewItem *>(item->itemAbove());
		}
		while(item && !item->isDir() && (item->parent() == currentSection));
	}

	setCurrentItem(lastItem);
	playCurrent();
	return current();
}

PlaylistItem Hayes::Playlist::getFirst(void) const
{
	if(!myBranch) return 0;
	return makePlaylistItem(getFirstItem(false, false));
}

PlaylistItem Hayes::Playlist::getLast(void) const
{
	if(!myBranch) return 0;
	return makePlaylistItem(getLastItem(false));
}

PlaylistItem Hayes::Playlist::getAfter(const PlaylistItem &prevItem) const
{
	if(!myBranch) return 0;
	FileTreeViewItem *item = findItem(prevItem);

	if(!item) return 0;
	return makePlaylistItem(getNextItem(item, false, false));
}

PlaylistItem Hayes::Playlist::getBefore(const PlaylistItem &nextItem) const
{
	if(!myBranch) return 0;
	FileTreeViewItem *item = findItem(nextItem);

	if(!item) return 0;
	return makePlaylistItem(getPreviousItem(item, false, false));
}

Hayes::FileTreeViewItem *Hayes::Playlist::viewItem(const PlaylistItem &pItem) const
{
	if(!myBranch) return 0;
	return findItem(pItem);
}

Hayes::FileTreeViewItem *Hayes::Playlist::getFirstItem(bool honorCheckBox, bool honorShuffle) const
{
	if(!myBranch || !myBranch->root()) return 0;

	myBranch->root()->sort();
	FileTreeViewItem *item = static_cast<FileTreeViewItem *>(myBranch->root());
	return getNextItem(item, honorCheckBox, honorShuffle);
}

Hayes::FileTreeViewItem *Hayes::Playlist::getLastItem(bool honorCheckBox) const
{
	if(!myBranch || !myBranch->root()) return 0;

	FileTreeViewItem *item = static_cast<FileTreeViewItem *>(myBranch->root());
	if(!item || !item->firstChild()) return 0;

	// Get the very very last item in the treeview
	// The, starting at that bottom, work up to the first valid item
	while(item->firstChild())
	{
		item = static_cast<FileTreeViewItem *>(item->firstChild());
		while(item->nextSibling()) item = static_cast<FileTreeViewItem *>(item->nextSibling());
	}

	if(!item || (!item->isDir() && (item->isOn() || !honorCheckBox)))
		return item;
	else
		return getPreviousItem(item, honorCheckBox, false);
}

Hayes::FileTreeViewItem *Hayes::Playlist::getNextItem(FileTreeViewItem *item,
                                                      bool honorCheckBox,
                                                      bool honorShuffle) const
{
	if(shuffle && honorShuffle)
		// Shuffle *always* honors the check boxes
		return getNextShuffleItem();

	if(!item) return getFirstItem(honorCheckBox, false);
	if(!item)
	{
		napp->player()->stop();
		return 0;
	}
	do
	{
		if(item->isDir())
		{
			if((item->isOn() || !honorCheckBox))
				openItem(item);
			else
				item->setOpen(false);
		}
		item = static_cast<FileTreeViewItem *>(item->itemBelow());
	}
	while(item && (item->isDir() || !(item->isOn() || !honorCheckBox)));

	return item;
}

Hayes::FileTreeViewItem *Hayes::Playlist::getPreviousItem(FileTreeViewItem *fitem,
                                                          bool honorCheckBox,
                                                          bool honorShuffle) const
{
	if(shuffle && honorShuffle)
		// Shuffle *always* honors the check boxes
		return getPreviousShuffleItem();

	if(!fitem) return 0;
	FileTreeViewItem *item = static_cast<FileTreeViewItem *>(fitem);
	do
	{
		item = static_cast<FileTreeViewItem *>(item->itemAbove());
		if(item && item->isDir())
		{
			if(item->isOn() || !honorCheckBox)
			{
				if(item == fitem->parent())
					fitem = item;
				else
				{
					// now go to the last child of the last directory
					QListViewItem *i;
					do 
					{
						openItem(item);
						if(!item->firstChild())
						{
							fitem = item;
							break;
						}
						for(i = item->firstChild(); i->nextSibling(); i = i->nextSibling());
						item = static_cast<FileTreeViewItem *>(i);
					}
					while(item->isDir());
				}
			}
			else
				item->setOpen(false);
		}
	}
	while(item && (item->isDir() || !(item->isOn() || !honorCheckBox)));

	return item;
}

namespace
{
// Shuffle helper
int weight(Hayes::FileTreeViewItem *item)
{
	if(!item || !item->isOn())
	{
		// easy
		return 0;
	}
	if(!item->isDir())
	{
		// Ordinary file
		return 1;
	}
	else if(!item->isOpen() && !item->childCount())
	{
		// a guess
		return 25;
	}
	else
	{
		int retval = 0;
		// Weight directory by contents
		for(QListViewItem *i = item->firstChild(); i; i = i->nextSibling())
			retval += weight(static_cast<Hayes::FileTreeViewItem *>(i));
		return retval;
	}
}
}

// Shuffle *always* honors the check boxes
Hayes::FileTreeViewItem *Hayes::Playlist::getNextShuffleItem(void) const
{
	// first check if there is a history to obey
	if(historyPosition != history.end() && ++historyPosition != history.end())
		return findItem(*historyPosition);

	// If no history, add to the end of it with a random item!
	// 1: start at root item
	FileTreeViewItem *root = static_cast<FileTreeViewItem *>(treeView->firstChild());
	// Careful...
	if(!root) return 0;

	FileTreeViewItem *item = root;
	openItem(item);
	// Careful...
	if(!item->firstChild()) return 0;

	// 2: pick a random checked child file/non-empty child dir
	//    of the directory
	do
	{
		QValueList<QListViewItem *> list;
		for(QListViewItem *i = item->firstChild(); i; i = i->nextSibling())
		{
			int count = weight(static_cast<FileTreeViewItem *>(i));
			// weight the items
			for(int j = 0; j < count; ++j)
				list.append(i);
		}
		// empty directory?  Go back to start
		if(!list.size())
			list.append(root);

		// Get a non-negative random number
		int randomNumber = -1;
		while(randomNumber < 0)
			randomNumber = kapp->random();

		// Weighted lookup!
		const int index = randomNumber % list.size();
		FileTreeViewItem *next = static_cast<FileTreeViewItem *>(list[index]);

		// Open the lucky item, if it is a directory
		if(next->isDir())
		{
			openItem(next);
			// Empty dirs are useless to us
			if(next->childCount())
				item = next;
			// Note that if childCount is 0, then item stays the same, so
			// we try again.
		}
		else
		{
			// Final file
			item = next;
		}
	}
	// 3: if item is a directory, goto 2
	while(item && item->isDir());

	// Careful...
	if(item)
	{
		history.append(item->fileItem()->url());
		historyPosition = history.end();
		--historyPosition;
	}
	return item;
}

// Shuffle *always* honors the check boxes
Hayes::FileTreeViewItem *Hayes::Playlist::getPreviousShuffleItem(void) const
{
	if(historyPosition == history.begin())
	{
		// if no history, start one
		if(historyPosition == history.end())
			return getNextShuffleItem();
		// otherwise loop to the end
		else
			historyPosition = history.end();
	}

	return findItem(*(--historyPosition));
}

// KFileTreeView may be asynchronous, but this isn't
void Hayes::Playlist::openItem(FileTreeViewItem *item) const
{
	//kdDebug() << "Opening " << item->fileItem()->url().prettyURL() << endl;
	if(!item->isDir() || item->isOpen()) return;
	treeView->touchItem(item);

	if(item->alreadyListed())
	{
		item->setOpen(true);
		return;
	}

	itemToOpen = item;
	QTimer::singleShot(0, const_cast<Playlist *>(this), SLOT(populateBegin(void)));
	kapp->enter_loop();
}

void Hayes::Playlist::populateBegin(void)
{
	connect(myBranch, SIGNAL(populateFinished(KFileTreeViewItem *)), this, SLOT(populateFinished(KFileTreeViewItem *)));
	emit busy(i18n("Opening %1").arg(itemToOpen->url().prettyURL()));
	itemToOpen->setOpen(true);
}

void Hayes::Playlist::populateFinished(KFileTreeViewItem *item)
{
	disconnect(myBranch, SIGNAL(populateFinished(KFileTreeViewItem *)), this, SLOT(populateFinished(KFileTreeViewItem *)));
	emit finished(i18n("Finished opening %1").arg(item->url().prettyURL()), 2000);
	item->sort();
	kapp->exit_loop();
	//kdDebug() << "Done opening " << item->fileItem()->url().prettyURL() << endl;
}

void Hayes::Playlist::openItemRecursively(FileTreeViewItem *item) const
{
	//kdDebug() << "Recursively opening " << item->fileItem()->url().prettyURL() << endl;
	openItem(item);

	for(QListViewItem *i = item->firstChild(); i; i = i->nextSibling())
	{
		FileTreeViewItem *fi = static_cast<FileTreeViewItem *>(i);
		if(fi->isDir())
			openItemRecursively(fi);
	}
	//kdDebug() << "Done recursively opening " << item->fileItem()->url().prettyURL() << endl;
}

void Hayes::Playlist::executed(QListViewItem *qitem)
{
	FileTreeViewItem *item = dynamic_cast<FileTreeViewItem *>(qitem);
	if(!item || item->isDir() || !item->supported()) return;

	setCurrentItem(item);
	playCurrent();
}

void Hayes::Playlist::itemDeleted(FileTreeViewItem *item)
{
	//kdDebug(66666) << "itemDeleted " << item->url().prettyURL() << endl;
	KURL url = item->url();
	for(QValueList<KURL>::Iterator i = history.begin(); i != history.end(); ++i)
	{
		if(*i == url || url.isParentOf(*i))
		{
			history.remove(i);
			if (historyPosition == i)
				historyPosition = history.end();
			i = history.begin();
		}
	}
	if(item == currentItem)
	{
		//kdDebug(66666) << "itemDeleted was current item" << endl;
		currentWasDeletedRudely();
	}
}

void Hayes::Playlist::cleared(void)
{
	myBranch = 0;
	currentWasDeletedRudely();
	clear();
}

void Hayes::Playlist::currentWasDeletedRudely(void)
{
	currentItem = 0;
	emit newSong(0);
	playCurrent();
}

void Hayes::Playlist::setCurrentItem(FileTreeViewItem *item)
{
	if(currentItem == item) return;

	//kdDebug(66666) << "setCurrentItem " << (void *)item << endl;

	if(saveVolume && currentItem)
		currentItem->setVolume(napp->player()->volume());

	currentItem = item;
	treeView->setSpecialItem(item);
	if(item)
	{
		treeView->ensureItemVisible(item);
		if(shuffle && (item->fileItem()->url() != *historyPosition))
		{
			history.append(item->fileItem()->url());
			historyPosition = history.end();
			--historyPosition;
		}

		if(saveVolume && item->hasVolume())
			napp->player()->setVolume(item->volume());
	}

	emit newSong(current());
}

PlaylistItem Hayes::Playlist::makePlaylistItem(FileTreeViewItem *item)
{
	if(!item) return 0;
	PlaylistItemData *data = new PlaylistItemData(*item->fileItem());
	return PlaylistItem(data);
}

Hayes::FileTreeViewItem *Hayes::Playlist::findItem(KURL url) const
{
	if(url == root) return static_cast<FileTreeViewItem *>(myBranch->root());
	if(!root.isParentOf(url)) return 0;

	// TODO: this *fails* when charsets are mismatched
	FileTreeViewItem *item = static_cast<FileTreeViewItem *>(treeView->findItem(myBranch,
	                                                         relativeString(root, url)));

	if(!item)
	{
		// Open parent of URL, so that url can actually be found
		KURL parentURL = url;
		parentURL.cd("..");
		FileTreeViewItem *parent = findItem(parentURL);
		if(!parent || !parent->isDir()) return 0;
		openItem(parent);

		// TODO: this *fails* when charsets are mismatched
		item = static_cast<FileTreeViewItem *>(treeView->findItem(myBranch,
		                                       relativeString(root, url)));
	}

	return item;
}

Hayes::FileTreeViewItem *Hayes::Playlist::findItem(PlaylistItem plItem) const
{
	// I don't trust Noatun as far as I can throw it
	const PlaylistItemData *d = dynamic_cast<const PlaylistItemData *>(plItem.data());
	if(!d) return 0; // Dork at the helm?

	// OK, now that the sanity check is passed
	return findItem(d->fileItem().url());
}

void Hayes::Playlist::turnOffAllChildren(const KURL &url)
{
	FileTreeViewItem *item = findItem(url);
	if(!item || !item->isDir()) return;
	openItem(item);
	for(QListViewItem *i = item->firstChild(); i; i = i->nextSibling())
	{
		FileTreeViewItem *fi = static_cast<FileTreeViewItem *>(i);
		fi->setOn(false);
	}
}

void Hayes::Playlist::turnOnAllChildren(const KURL &url)
{
	FileTreeViewItem *item = findItem(url);
	if(!item || !item->isDir()) return;
	openItem(item);
	for(QListViewItem *i = item->firstChild(); i; i = i->nextSibling())
	{
		FileTreeViewItem *fi = static_cast<FileTreeViewItem *>(i);
		fi->setOn(true);
	}
}

void Hayes::Playlist::openAllChildren(const KURL &url)
{
	if(url == root)
	{
		expandAll();
		return;
	}

	FileTreeViewItem *item = findItem(url);
	if(!item || !item->isDir()) return;
	openItem(item);
	for(QListViewItem *i = item->firstChild(); i; i = i->nextSibling())
	{
		FileTreeViewItem *fi = static_cast<FileTreeViewItem *>(i);
		openItemRecursively(fi);
	}
}

void Hayes::Playlist::closeAllChildren(const KURL &url)
{
	if(url == root)
	{
		collapseAll();
		return;
	}

	FileTreeViewItem *item = findItem(url);
	if(!item || !item->isDir()) return;
	for(QListViewItem *i = item->firstChild(); i; i = i->nextSibling())
	{
		FileTreeViewItem *fi = static_cast<FileTreeViewItem *>(i);
		FileTreeView::closeItemRecursively(fi);
	}
	item->setOpen(false);
}

void Hayes::Playlist::expandAll(void)
{
	if(!myBranch) return;

	openItemRecursively(static_cast<FileTreeViewItem *>(myBranch->root()));
}

void Hayes::Playlist::collapseAll()
{
	if(!myBranch) return;

	FileTreeView::closeItemRecursively(static_cast<FileTreeViewItem *>(myBranch->root()));
}

#include "playlist.moc"
// arch-tag: playlist.cpp
