// Copyright (C) 2002 Neil Stevens <neil@hakubi.us>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
// AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name(s) of the author(s) shall not be
// used in advertising or otherwise to promote the sale, use or other dealings
// in this Software without prior written authorization from the author(s).

#ifndef HAYES_PLAYLIST_H
#define HAYES_PLAYLIST_H

#include <kfiletreeview.h>
#include <kfileitem.h>
#include <kurl.h>
#include <noatun/playlist.h>
#include <qvaluestack.h>
#include <qwidget.h>

#include "branch.h"
#include "dcopinterface.h"

class KPopupMenu;

namespace Hayes
{
// This class is tricky, as I have to ignore most of Noatun's old
// tag loading system, providing instead the great new KFileMetaInfo stuff.
//
// Though at the same time all the info must be available
class PlaylistItemData : public ::PlaylistItemData
{
public:
	PlaylistItemData(const KFileItem &);
	virtual ~PlaylistItemData(void);

	virtual QString property(const QString &, const QString & = 0) const;
	virtual void setProperty(const QString &, const QString &);
	virtual void clearProperty(const QString &);
	virtual QStringList properties(void) const;
	virtual bool isProperty(const QString &) const;
	virtual bool operator==(const PlaylistItemData &) const;

	const KFileItem &fileItem(void) const;

	virtual void remove(void) {};
private:
	KURL url;
	KFileItem *item;
	int len;
	QDict<QString> *dict;
};

// This is an entirely filesystem-based playlist.
//
// It supplies you a view, because the view has the DirLister, and the Playlist
// needs the DirLister.
class Playlist : public ::Playlist, virtual public DCOPInterface
{
Q_OBJECT
public:
	Playlist(QWidget *viewParent, QWidget *listParent, const char * = "HayesView", const char * = "HayesPlaylist");

	FileTreeView *view(void) { return treeView; };

	void open(const KURL &);
	virtual KURL playlistDirectory(void) const;

	void setColumns(const int &a) { treeView->setColumns(a); };
	void sortDirectoriesFirst(const bool &b) { treeView->sortDirectoriesFirst(b); };
	void caseSensitiveSort(const bool &b) { treeView->caseSensitiveSort(b); };
	void hideDotFiles(const bool &b) { treeView->hideDotFiles(b); };
	void hideUnplayableFiles(const bool &b) { treeView->hideUnplayableFiles(b); };

	virtual void setShuffle(bool);
	virtual void toggleShuffle(void);
	virtual bool isShuffling(void) const;

	virtual void expandAll(void);
	virtual void collapseAll(void);

	virtual void setSaveVolume(bool);
	virtual void toggleSaveVolume(void);
	virtual bool isSavingVolume(void) const;

	virtual void reset(void);
	virtual void clear(void);
	virtual void setCurrent(const PlaylistItem &);
	virtual void setCurrent(const KURL &);
	virtual void setCurrent(const QString &);
	virtual void addFile(const KURL &, bool = false) {};
	virtual PlaylistItem next(void);
	virtual PlaylistItem current(void);
	virtual PlaylistItem previous(void);
	virtual PlaylistItem getFirst(void) const;
	virtual PlaylistItem getLast(void) const;
	virtual PlaylistItem getAfter(const PlaylistItem &) const;
	virtual PlaylistItem getBefore(const PlaylistItem &) const;

	FileTreeViewItem *viewItem(const PlaylistItem &) const;

	virtual bool listVisible(void) const
	{
		return ((QWidget *)parent())->isVisible();
	}

	void turnOffAllChildren(const KURL &);
	void turnOnAllChildren(const KURL &);
	void openAllChildren(const KURL &);
	void closeAllChildren(const KURL &);

public slots:
	// Slot to handle reopen from TreeView
	void clearAndReopen(void);

	void cleared(void);
	virtual PlaylistItem nextSection(void);
	virtual PlaylistItem previousSection(void);

	// hmm.. but I thought C casts were bad, mmkay?
	// I hope I don't end up in a ditch somewhere because of this.
	virtual void showList(void)
	{
		if(!listVisible())
		{
			((QWidget *)parent())->show();
			((QWidget *)parent())->raise();
		}
	}

	// This is just one little itsy bitsy area where C++ strong typing
	// is a nuisance.
	virtual void hideList(void)
	{
		if(listVisible())
		{
			((QWidget *)parent())->hide();
		}
	}

signals:
	void busy(const QString &);
	void finished(const QString &, int);
	void newSong(PlaylistItem);

	void shuffleChanged(bool);
	void saveVolumeChanged(bool);

private slots:
	void populateBegin(void);
	void populateFinished(KFileTreeViewItem *);
	void executed(QListViewItem *);

	void itemDeleted(FileTreeViewItem *);
private:
	FileTreeViewItem *getFirstItem(bool honorCheckBox, bool honorShuffle) const;
	FileTreeViewItem *getLastItem(bool honorCheckBox) const;
	FileTreeViewItem *getNextItem(FileTreeViewItem *, bool honorCheckBox, bool honorShuffle) const;
	FileTreeViewItem *getPreviousItem(FileTreeViewItem *, bool honorCheckBox, bool honorShuffle) const;

	void openItem(FileTreeViewItem *) const;
	void openItemRecursively(FileTreeViewItem *) const;

	void setCurrentItem(FileTreeViewItem *);
	void currentWasDeletedRudely(void);

	static PlaylistItem makePlaylistItem(FileTreeViewItem *);
	FileTreeViewItem *findItem(KURL) const;
	FileTreeViewItem *findItem(PlaylistItem) const;

	KURL root;
	FileTreeView *treeView;
	mutable Branch *myBranch;
	FileTreeViewItem *currentItem;
	mutable FileTreeViewItem *itemToOpen;

	bool saveVolume;

	bool shuffle;
	FileTreeViewItem *getNextShuffleItem(void) const;
	FileTreeViewItem *getPreviousShuffleItem(void) const;
	// By popular demand, Shuffle has a history
	typedef QValueStack<KURL> History;
	// getAfter and getBefore are const, so to avoid having to have
	// two copies of the same get*Items, these must be mutable.
	mutable History history;
	mutable History::Iterator historyPosition;
};

}

#endif
// arch-tag: playlist.h
