// Copyright (C) 2002 Neil Stevens <neil@hakubi.us>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
// AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name(s) of the author(s) shall not be
// used in advertising or otherwise to promote the sale, use or other dealings
// in this Software without prior written authorization from the author(s).

#include <cassert>
#include <kaboutapplication.h>
#include <kaboutdata.h>
#include <kaction.h>
#include <kconfig.h>
#include <kdebug.h>
#include <kdialog.h>
#include <keditcl.h>
#include <kedittoolbar.h>
#include <kfiledialog.h>
#include <kglobal.h>
#include <kkeydialog.h>
#include <klocale.h>
#include <kmenubar.h>
#include <kmessagebox.h>
#include <kpopupmenu.h>
#include <kpropertiesdialog.h>
#include <kstatusbar.h>
#include <kstdaction.h>
#include <kstandarddirs.h>
#include <kurldrag.h>
#include <noatun/app.h>
#include <noatun/player.h>
#include <noatun/stdaction.h>
#include <qfileinfo.h>
#include <qframe.h>
#include <qheader.h>
#include <qlayout.h>
#include <qtooltip.h>

#include "cmodule.h"
#include "playlist.h"
#include "propertypage.h"
#include "shell.h"
#include "window.h"

extern "C" Plugin *create_plugin()
{
	return new Hayes::Window;
}

Hayes::Window::Window(void)
	: KMainWindow(0, "Hayes")
	, Plugin()
	, shell(0)
	, dirMenu(new KPopupMenu(this))
	, menu(new KPopupMenu(this))
	, contextItem(0)
{
	setStandardToolBarMenuEnabled(true);

	list = new Hayes::Playlist(this, this);
	connect(this, SIGNAL(listHidden()), list, SIGNAL(listHidden()));
	connect(this, SIGNAL(listShown()), list, SIGNAL(listShown()));

	(void)KStdAction::open(this, SLOT(file_open()), actionCollection());
	(void)KStdAction::close(this, SLOT(close()), actionCollection());
	(void)KStdAction::find(this, SLOT(edit_find()), actionCollection());
	options_menubarAction = KStdAction::showMenubar(this, SLOT(options_menubar()), actionCollection());
	options_statusbarAction = KStdAction::showStatusbar(this, SLOT(options_statusbar()), actionCollection());
	options_shuffleAction = new KToggleAction(i18n("&Shuffle"), 0, this, SLOT(options_shuffle()), actionCollection(), "options_shuffle");
	options_volumeAction = new KToggleAction(i18n("Save &Volume"), 0, this, SLOT(options_volume()), actionCollection(), "options_volume");
	(void)KStdAction::preferences(napp, SLOT(preferences()), actionCollection());
	(void)KStdAction::configureToolbars(this, SLOT(options_configuretoolbars() ), actionCollection());
	(void)KStdAction::keyBindings(this, SLOT(options_configurekeys()), actionCollection());
	NoatunStdAction::back(actionCollection(), "back");
	NoatunStdAction::stop(actionCollection(), "stop");
	NoatunStdAction::play(actionCollection(), "play");
	NoatunStdAction::pause(actionCollection(), "pause");
	NoatunStdAction::forward(actionCollection(), "forward");
	NoatunStdAction::effects(actionCollection(), "effects");
	NoatunStdAction::equalizer(actionCollection(), "equalizer");
	(void)new KAction(i18n("Previous Section"), 0, list, SLOT(previousSection(void)), actionCollection(), "previoussection");
	(void)new KAction(i18n("Next Section"), 0, list, SLOT(nextSection(void)), actionCollection(), "nextsection");
	(void)new KAction(i18n("About Hayes..."), 0, this, SLOT(aboutHayes(void)), actionCollection(), "abouthayes");
	(void)new KAction(i18n("Configure Hayes..."), 0, this, SLOT(configureHayes(void)), actionCollection(), "configurehayes");

	seekSlider = new L33tSlider(0, 1000, 10, 0, L33tSlider::Horizontal, this, "seek-slider");
	seekSlider->setOrientation(Horizontal);
	connect(seekSlider, SIGNAL(userChanged(int)), SLOT(seekSliderChanged(int)));
	initSeekSlider();
	(void)new KWidgetAction(seekSlider, i18n("Seek"), KShortcut(), 0, 0, actionCollection(), "seek");

	volumeSlider = new L33tSlider(0, "volume-slider");
	volumeSlider->setOrientation(Horizontal);
	connect(volumeSlider, SIGNAL(userChanged(int)), SLOT(volumeSliderChanged(int)));
	initVolumeSlider();
	(void)new KWidgetAction(volumeSlider, i18n("Volume"), KShortcut(), 0, 0, actionCollection(), "volume");

	createGUI("hayesui.rc");
	menuBar()->insertItem(i18n("&Actions"), napp->pluginMenu(), -1, 3);

	currentSong = new QLabel(statusBar());
	statusBar()->addWidget(currentSong,1);
	statusBar()->show();
	newSong(0);

	connect(list, SIGNAL(busy(const QString &)), statusBar(), SLOT(message(const QString &)));
	connect(list, SIGNAL(finished(const QString &, int)), statusBar(), SLOT(message(const QString &, int)));
	connect(list, SIGNAL(newSong(PlaylistItem)), this, SLOT(newSong(PlaylistItem)));
	connect(list, SIGNAL(shuffleChanged(bool)), options_shuffleAction, SLOT(setChecked(bool)));
	connect(list, SIGNAL(saveVolumeChanged(bool)), options_volumeAction, SLOT(setChecked(bool)));

	connect(list->view(), SIGNAL(contextMenu(KListView *, QListViewItem *, const QPoint &)), this, SLOT(contextMenu(KListView *, QListViewItem *, const QPoint &)));
	menu->insertItem(i18n("Properties..."), this, SLOT(properties(void)));
	dirMenu->insertItem(i18n("Properties..."), this, SLOT(properties(void)));
	dirMenu->insertItem(i18n("Open in Browser"), this, SLOT(openInBrowser(void)));
	dirMenu->insertSeparator();
	dirMenu->insertItem(i18n("Open All Children"), this, SLOT(openAllChildren(void)));
	dirMenu->insertItem(i18n("Close All Children"), this, SLOT(closeAllChildren(void)));
	dirMenu->insertSeparator();
	dirMenu->insertItem(i18n("Turn Off All Children"), this, SLOT(turnOffAllChildren(void)));
	dirMenu->insertItem(i18n("Turn On All Children"), this, SLOT(turnOnAllChildren(void)));

	connect(napp->player(), SIGNAL(timeout(void)), this, SLOT(updateSeekSlider(void)));
	connect(napp->player(), SIGNAL(newSong(void)), this, SLOT(initSeekSlider(void)));
	connect(napp->player(), SIGNAL(stopped(void)), this, SLOT(resetSeekSlider(void)));

	connect(napp->player(), SIGNAL(timeout(void)), this, SLOT(updateVolumeSlider(void)));
	connect(napp->player(), SIGNAL(newSong(void)), this, SLOT(updateVolumeSlider(void)));
	connect(napp->player(), SIGNAL(stopped(void)), this, SLOT(updateVolumeSlider(void)));
}

Hayes::Window::~Window()
{
	KConfig &config = *KGlobal::config();
	saveMainWindowSettings(&config, "Hayes Window");
	config.setGroup("Hayes");
	config.writeEntry("currentPlaylistURL", currentPlaylistURL.url(-1));
	config.writeEntry("shuffle", options_shuffleAction->isChecked());
	config.writeEntry("volume", options_volumeAction->isChecked());

	list->view()->superSaveLayout(&config, "Hayes ListView");

	delete list;
	list = 0;
	delete shell;
	shell = 0;
}

void Hayes::Window::init(void)
{
	setCentralWidget(list->view());

	hide();

	// this has to come after the load, because of the modified flag
	KConfig &config = *KGlobal::config();

	applyMainWindowSettings(&config, "Hayes Window");
	options_statusbarAction->setChecked( !statusBar()->isHidden() );
	options_menubarAction->setChecked( !menuBar()->isHidden() );

	currentPlaylistURL = KURL();
	config.setGroup("Hayes");

	options_shuffleAction->setChecked(config.readBoolEntry("shuffle", false));
	options_shuffle();

	options_volumeAction->setChecked(config.readBoolEntry("volume", false));
	options_volume();

	list->view()->superRestoreLayout(&config, "Hayes ListView");

	Hayes::CModule *module = new Hayes::CModule(this);
	module->reopen();
	module->save();

	// this must be last, specificly the open().  The open() enters a new loop
	QString playlistURL = config.readEntry("currentPlaylistURL");
	if(!playlistURL.isEmpty())
	{
		currentPlaylistURL = KURL(playlistURL);
		setCaption(currentPlaylistURL.prettyURL());
		list->open(currentPlaylistURL);
	}
	else
	{
		setCaption(i18n("Playlist"));
	}
}

void Hayes::Window::seekSliderChanged(int value)
{
	napp->player()->skipTo(value * 1000);

	QToolTip::add(seekSlider, i18n("Seek: %1").arg( napp->player()->lengthString()));
}

void Hayes::Window::initSeekSlider()
{
	int length = napp->player()->getLength();
	seekSlider->setRange(0, kMax(0, length));
	seekSlider->setEnabled(length > 0);
	seekSlider->setValue(0);
	seekSlider->setEnabled(true);
	QToolTip::add(seekSlider, i18n("Seek: %1").arg( napp->player()->lengthString()));
}

void Hayes::Window::updateSeekSlider()
{
	if (!napp->player()->isStopped())
	{
		seekSlider->setRange(0, napp->player()->getLength() / 1000);
		seekSlider->setValue(napp->player()->getTime() / 1000);
	}

	QToolTip::add(seekSlider, i18n("Seek: %1").arg( napp->player()->lengthString()));
}

void Hayes::Window::resetSeekSlider()
{
	seekSlider->setValue(0);
	seekSlider->setEnabled(false);
}

void Hayes::Window::volumeSliderChanged(int value)
{
	napp->player()->setVolume(value);

	QToolTip::add(volumeSlider, i18n("Volume: %1").arg( napp->player()->volume()));
}

void Hayes::Window::initVolumeSlider()
{
	volumeSlider->setRange(0, 100);
	volumeSlider->setValue(napp->player()->volume());
	volumeSlider->setEnabled(true);
	QToolTip::add(volumeSlider, i18n("Volume: %1").arg( napp->player()->volume()));
}

void Hayes::Window::updateVolumeSlider()
{
	volumeSlider->setValue(napp->player()->volume());

	QToolTip::add(volumeSlider, i18n("Volume: %1").arg( napp->player()->volume()));
}

Playlist *Hayes::Window::playlist(void)
{
	return list;
}

void Hayes::Window::closeEvent(QCloseEvent *)
{
	if(list) list->hideList();
	hide();

	// XXX destructor isn't being called?? so save config here just to be safe...
	KConfig &config = *KGlobal::config();
	saveMainWindowSettings(&config, "Hayes Window");
	config.setGroup("Hayes");
	config.writeEntry("currentPlaylistURL", currentPlaylistURL.url(-1));
	config.writeEntry("shuffle", options_shuffleAction->isChecked());
	config.writeEntry("volume", options_volumeAction->isChecked());

	list->view()->superSaveLayout(&config, "Hayes ListView");
}

void Hayes::Window::hideEvent(QHideEvent *e)
{
	KMainWindow::hideEvent(e);
	emit listHidden();
}

void Hayes::Window::showEvent(QShowEvent *e)
{
	KMainWindow::showEvent(e);
	emit listShown();
}

void Hayes::Window::search(void)
{
	bool done = false;
	bool found = false;

	while(!done)
	{
		if(!findCurrent)
		{
			if(findDialog->get_direction())
			{
				// backward
				if(KMessageBox::Yes == KMessageBox::questionYesNo(findDialog,
				   i18n("Beginning of document reached.\nContinue from the end?"),
				   i18n("Find"))
				)
				{
					findCurrent = list->getFirst();
				}
				else
				{
					findCurrent = list->getLast();
					done = true;
				}
			}
			else
			{
				// forward
				if(KMessageBox::Yes == KMessageBox::questionYesNo(findDialog,
				   i18n("End of document reached.\nContinue from the beginning?"),
				   i18n("Find"))
				)
				{
					findCurrent = list->getFirst();
				}
				else
				{
					findCurrent = list->getLast();
					done = true;
				}
			}
		}

		kdDebug(66666) << "find: " << findCurrent.title() << endl;
		KFileTreeViewItem *item = list->viewItem(findCurrent);
		assert(item);
		for(int i = 0; i < list->view()->columns() && !found; i++)
		{
			if(item->text(i).contains(findDialog->getText(), findDialog->case_sensitive()))
			{
				list->view()->clearSelection();
				list->view()->setSelected(item, true);
				list->view()->ensureItemVisible(item);
				found = true;
				done = true;
			}
		}

		if(!findDialog->get_direction())
			findCurrent = list->getAfter(findCurrent); // forward
		else
			findCurrent = list->getBefore(findCurrent); // backward
	}

	if(!found) KMessageBox::information(findDialog, i18n("Text not found."), i18n("Find"));
}

void Hayes::Window::searchDone(void)
{
	napp->exit_loop();
}

void Hayes::Window::newSong(PlaylistItem item)
{
	QString cap = i18n("Not playing");
	if(item) cap = item.title();
	currentSong->setText(cap);
}

void Hayes::Window::contextMenu(KListView *, QListViewItem *qitem, const QPoint &point)
{
	contextItem = dynamic_cast<KFileTreeViewItem *>(qitem);
	if(!contextItem) return;
	if(contextItem->isDir())
		dirMenu->popup(point);
	else
		menu->popup(point);
}

void Hayes::Window::properties(void)
{
	assert(contextItem);
	KPropertiesDialog *dialog = new KPropertiesDialog(contextItem->fileItem());
	// removed until it has something in it
	// dialog->insertPlugin(new PropertyPage(dialog));
}

void Hayes::Window::openInBrowser(void)
{
	assert(contextItem);
	kapp->invokeBrowser(contextItem->fileItem()->url().url());
}

void Hayes::Window::turnOffAllChildren(void)
{
	assert(contextItem);
	list->turnOffAllChildren(contextItem->fileItem()->url());
}

void Hayes::Window::turnOnAllChildren(void)
{
	assert(contextItem);
	list->turnOnAllChildren(contextItem->fileItem()->url());
}

void Hayes::Window::openAllChildren(void)
{
	assert(contextItem);
	list->openAllChildren(contextItem->fileItem()->url());
}

void Hayes::Window::closeAllChildren(void)
{
	assert(contextItem);
	list->closeAllChildren(contextItem->fileItem()->url());
}

void Hayes::Window::file_open(void)
{
	KURL url = KFileDialog::getExistingDirectory();
	if(!url.isEmpty())
	{
		currentPlaylistURL = url;
		list->open(currentPlaylistURL);
		setCaption(currentPlaylistURL.prettyURL());

		// save here for more predictable behavior
		KConfig &config = *KGlobal::config();
		config.setGroup("Hayes");
		config.writeEntry("currentPlaylistURL", currentPlaylistURL.url(-1));
		config.sync();
	}
}

void Hayes::Window::edit_find(void)
{
	findCurrent = list->current();
	if(!findCurrent) findCurrent = list->getFirst();

	findDialog = new KEdFind(this);
	connect(findDialog, SIGNAL(search()), this, SLOT(search()));
	connect(findDialog, SIGNAL(done()), this, SLOT(searchDone()));

	// enter loop
	findDialog->exec();

	// after loop is done
	delete findDialog;
}

void Hayes::Window::options_shuffle(void)
{
	list->setShuffle(options_shuffleAction->isChecked());
}

void Hayes::Window::options_volume(void)
{
	list->setSaveVolume(options_volumeAction->isChecked());
}

void Hayes::Window::options_menubar(void)
{
	if(options_menubarAction->isChecked())
		menuBar()->show();
	else
		menuBar()->hide();
}

void Hayes::Window::options_statusbar(void)
{
	if(options_statusbarAction->isChecked())
		statusBar()->show();
	else
		statusBar()->hide();
}

void Hayes::Window::options_configuretoolbars(void)
{
	KEditToolbar dlg(actionCollection(), "hayesui.rc");
	if(dlg.exec())
		createGUI("hayesui.rc");
}

void Hayes::Window::options_configurekeys(void)
{
	KKeyDialog::configureKeys(actionCollection(), xmlFile());
}

void Hayes::Window::aboutHayes(void)
{
const char *license = "Copyright (C) 2002 Neil Stevens <neil@hakubi.us>\n\
\n\
Permission is hereby granted, free of charge, to any person obtaining a copy\n\
of this software and associated documentation files (the ""Software""), to deal\n\
in the Software without restriction, including without limitation the rights\n\
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell\n\
copies of the Software, and to permit persons to whom the Software is\n\
furnished to do so, subject to the following conditions:\n\
\n\
The above copyright notice and this permission notice shall be included in\n\
all copies or substantial portions of the Software.\n\
\n\
THE SOFTWARE IS PROVIDED ""AS IS"", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\n\
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\n\
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE\n\
THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN\n\
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN\n\
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.\n\
\n\
Except as contained in this notice, the name(s) of the author(s) shall not be\n\
used in advertising or otherwise to promote the sale, use or other dealings\n\
in this Software without prior written authorization from the author(s).";

	KAboutData data("hayes", I18N_NOOP("Hayes"), "1.5",
	               I18N_NOOP("Filesystem-based Playlist"),
	               KAboutData::License_Custom,
	               "Copyright (C) 2002, 2003 Neil Stevens", 0,
	               "http://www.hakubi.us/hayes/", "bugs@hakubi.us");
	data.addAuthor("Neil Stevens", I18N_NOOP("Author"), "neil@hakubi.us");
	data.addCredit("Carsten Pfeiffer", I18N_NOOP("Optimizations, Slider"), "pfeiffer@kde.org");
	data.addCredit("Rob Kaper", I18N_NOOP("The Hero of Hayes"), "cap@capsi.com", "http://unixcode.org/");
	data.addCredit("James Stembridge", I18N_NOOP("Fixes"));
	data.addCredit("\"Spoonz\"", I18N_NOOP("Volume Slider, Collapse All Items action"));
	data.addCredit("United States Army", I18N_NOOP("Preserving the freedom that made this software possible"));
	data.setLicenseText(license);

	KAboutApplication dialog(&data, this);
	dialog.exec();
}

void Hayes::Window::configureHayes(void)
{
	if(!shell)
	{
		shell = new Shell(0, "Hayes::Shell", false);
		shell->setModule(new Hayes::Module(shell));
	}
	shell->show();
}

#include "window.moc"
// arch-tag: window.cpp
